<?php

if (!defined('DIAFAN')) {
    $path = __FILE__;
    $i    = 0;
    while (!file_exists($path . '/includes/404.php')) {
        if ($i == 10) {
            exit;
        }
        $i++;
        $path = dirname($path);
    }
    include $path . '/includes/404.php';
}

class OfdfermaException extends Exception{}

/**
 * @property string inn
 * @property string taxation_system
 * @property string vat
 * @property string vat_delivery
 * @property string public_id
 * @property string secret_key
 */
class Ofdferma_inc extends Model {

    const MODULE_NAME = 'ofdferma';
    
    const API_URL   = 'https://ferma.ofd.ru/api/kkt/cloud';
    const AUTH_URL  = 'https://ferma.ofd.ru/api/Authorization/CreateAuthToken';
    const CHECK_URL = 'https://check.ofd.ru/rec/';
    
    static private $ofdferma_token = null;
    static private $ofdferma_token_exp_date = null;
    
    private $_debug = 0;
    
    private function checkEnabled(){
        if ( ! $this->diafan->configmodules('enabled', 'ofdferma')) {
            return false;
        }
        
        return true;
    }
    
    public function getDataStatus(){
        return DB::query_fetch_all("SELECT id, [name] FROM {shop_order_status}");
    }
    
    public function getDataNds(){
        return array(
            array(
                'id'   => 'Vat0',
                'name' => 'Без НДС',
            ),
            array(
                'id'   => 'Vat10',
                'name' => '10%',
            ),
            array(
                'id'   => 'Vat18',
                'name' => '18%',
            ),
			array(
                'id'   => 'Vat20',
                'name' => '20%',
            ),
            array(
                'id'   => 'CalculatedVat10110',
                'name' => '10/110',
            ),
            array(
                'id'   => 'CalculatedVat18118',
                'name' => '18/118',
            ),
			array(
                'id'   => 'CalculatedVat20120',
                'name' => '20/120',
            ),
        );
    }
    
    private function checkNds($val){
        foreach($this -> getDataNds() as $aNdsItem){
            if ($aNdsItem['id'] == $val){
                return true;
                
                break;
            }
        }
        
        return FALSE;
    }
    
    public function set_tovar_nds($tovar_id, $val){
        DB::query("DELETE FROM {ofdferma_tovars} WHERE tovar_id = %d", $tovar_id);
        DB::query("INSERT INTO {ofdferma_tovars} (tovar_id, nds_id) VALUES (%d, '%s')", $tovar_id, $val);
    }
    
    public function get_tovar_nds($tovar_id){
        $var = DB::query_fetch_value("SELECT nds_id FROM {ofdferma_tovars} WHERE tovar_id = %d", $tovar_id, "nds_id");
        
        return isset($var[0]) ? $var[0] : '';
    }
    
    public function set_status_order($order, $status){
        if ( ! $this -> checkEnabled()){
            return;
        }
        
        //Если статус при котором формировать чек
        if ( $status['id'] == $this->diafan->configmodules('status', 'ofdferma')) {
            $this -> setDebug(1);
            $this -> check($order['id'], 'Income');
            $this -> setDebug(0);
        }
    }

    private function get_receipt($order_id, $type)
    {
        return DB::query_fetch_array("SELECT * FROM {ofdferma_receipt} WHERE order_id=%d && type='%s' LIMIT 1", $order_id, $type);
    }
    
    public function update_receipt($data, $id){
        DB::query(sprintf("
            UPDATE {ofdferma_receipt} SET
            status='%s',
            status_message='%s',
            FN='%s',
            RNM='%s',
            FDN='%s',
            FPD='%s',
            updated_at='%s'
            WHERE id='%s'
        ", $data['status'], $data['status_message'], $data['FN'], $data['RNM'], $data['FDN'], $data['FPD'], gmdate('Y-m-d H:i:s'), 
            $id
        ));
    }
    
    public function insert_receipt($data){
        DB::query(sprintf("
            INSERT INTO {ofdferma_receipt} SET
            id='%s',
            type='%s',
            inn='%s',
            order_id=%d,
            total=%d,
            created_at='%s'
        ", $data['id'], $data['type'], $data['inn'], $data['order_id'], $data['total'], gmdate('Y-m-d H:i:s')));
    }
    
    /*====================================*/
    private function getOrder($orderID)
    {
        return DB::query_fetch_array("SELECT * FROM {shop_order} WHERE id=%d AND trash='0'", $orderID);
    }

    /**
     * Возвращает E-Mail пользователя, совершившего заказ
     *
     * @param int $orderID Идентификатор заказа
     *
     * @return mixed
     */
    private function getOrderField($orderID, $type = '')
    {
        $result = DB::query_result("
            SELECT value FROM {shop_order_param_element} AS e
            INNER JOIN {shop_order_param} AS p ON p.id=e.param_id AND p.type='%s'
            WHERE e.element_id=%d ", $type, $orderID);
        return $result ? $result : '';
    }

    private function getOrderPositions($order)
    {
        $nds_global = $this->diafan->configmodules('nds', 'ofdferma');
        
        
        $result = array();

        $summ_goods = 0;

        $rows = DB::query_fetch_all("
            SELECT g.id, g.count_goods AS `count`, g.good_id, g.price, g.discount_id, s.[name]
            FROM {shop_order_goods} AS g
            INNER JOIN {shop} AS s ON g.good_id=s.id
            WHERE g.order_id=%d ORDER by g.id ASC
        ", $order['id']);

        $ids = array();
        foreach ($rows as $row) {
            $ids[] = $row["id"];
        }

        $good_params = DB::query_fetch_key_array("SELECT * FROM {shop_order_goods_param} WHERE order_goods_id IN (%s)", implode(",", $ids), "order_goods_id");
        $discount_ids = array();
        foreach ($rows as $i => $row) {
            $params = array();
            if (!empty($good_params[$row["id"]])) {
                foreach ($good_params[$row["id"]] as $row_p) {
                    $params[$row_p["param_id"]] = $row_p["value"];
                }
            }
            $row_price = $this->diafan->_shop->price_get($row["good_id"], $params, false);
            if ($row_price) {
                $row_prices[$row["id"]] = $row_price;
            }
            if ($row["discount_id"] && !in_array($row["discount_id"], $discount_ids)) {
                $discount_ids[] = $row["discount_id"];
            }
        }

        if (!empty($discount_ids)) {
            $discounts = DB::query_fetch_key("SELECT id, discount, deduction FROM {shop_discount} WHERE id IN (%s)", implode(",", $discount_ids), "id");
        } else {
            $discounts = array();
        }

        $additional_costs = DB::query_fetch_key_array("
            SELECT a.id, a.[name], s.summ, s.order_goods_id FROM {shop_additional_cost} AS a
            INNER JOIN {shop_order_additional_cost} AS s ON s.additional_cost_id=a.id AND s.order_id=%d
            WHERE a.trash='0' AND a.shop_rel='1'
            ORDER BY a.sort ASC
        ", $order['id'], "order_goods_id");

        if ($order['discount_summ']) {
            $total_rows = count($rows);
            $order_discount_part = floatval($order['discount_summ']) / $total_rows;
        } else {
            $order_discount_part = 0;
        }

        foreach ($rows as $row) {
            $row_price = !empty($row_prices[$row["id"]]) ? $row_prices[$row["id"]] : false;
            if (!empty($additional_costs[$row["id"]])) {
                foreach ($additional_costs[$row["id"]] as $a) {
                    $row["price"] += $a["summ"] / $row["count"];
                }
            }

            $row["summ"] = $row["count"] * $row["price"];

            $summ_goods += $row["summ"];

            if ($row["discount_id"] && !empty($discounts[$row["discount_id"]])) {
                $discount = $discounts[$row["discount_id"]];
                if (!empty($discount["deduction"])) {
                    $row["discount"] = $discount["deduction"];
                } else {
                    $row["discount"] = $row["price"] / (100 - $discount["discount"]) * $discount["discount"];
                }
            } elseif (!empty($row_price["old_price"])) {
                $row["discount"] = $row_price["old_price"] - $row["price"];
            } else {
                $row["discount"] = 0;
            }

            $row['discount'] += $order_discount_part;

            $nds = $this -> get_tovar_nds($row['good_id']);            
                    
            $result[] = array(
                'name'      => $row['name'],
                'price'     => (float) $row['price'],
                'count'     => (float) $row['count'],
                'total'     => floatval($row['summ'] - $row['discount']),
                'discount'  => (float) $row['discount'],
                'vat'       => $nds
            );
        }

        if ($order["discount_summ"]) {
            $summ_goods = $summ_goods - $order["discount_summ"];
        }

        $additional_costs = DB::query_fetch_all("
            SELECT
                a.id, a.[name], a.price, a.percent, a.[text], a.amount, a.required, o.summ
            FROM {shop_additional_cost} AS a
            INNER JOIN {shop_order_additional_cost} AS o ON o.additional_cost_id=a.id
            WHERE a.trash='0' AND o.order_id=%d AND a.shop_rel='0' ORDER by sort ASC
        ", $order['id']);

        foreach ($additional_costs as $row) {
            if ($row['summ']) {
                $row['price'] = $row['summ'];
            } else {
                if (!empty($row['amount']) && $row['amount'] < $summ_goods) {
                    // Услуга предоставляется бесплатно для суммы > $row['amount']
                    $row['summ'] = 0;
                } elseif ($row['percent']) {
                    $row['summ'] = $summ_goods * $row['percent'] / 100;
                    $row['price'] = $row['summ'];
                } else {
                    $row['summ'] = $row['price'];
                }
            }
            
            $result[] = array(
                'name'      => $row['name'], 
                'price'     => (float) $row['price'], 
                'count'     => 1, 
                'total'     => (float) $row['summ'], 
                'discount'  => 0, 
                'vat'       => $nds_global
            );
        }

        if ($order['delivery_summ']) {
            $delivery_summ = (float) $order['delivery_summ'];
            $result[] = array(
                'name'      => 'Доставка', 
                'price'     => $delivery_summ, 
                'count'     => 1, 
                'total'     => $delivery_summ, 
                'discount'  => 0, 
                'vat'       => $nds_global
            );
        }

        return $result;
    }
    
    /*===================================*/
    private function check_settings() {
        $ofd_login      = $this->diafan->configmodules('login', 'ofdferma');
        $ofd_pass       = $this->diafan->configmodules('password', 'ofdferma');
        $ofd_nalog      = $this->diafan->configmodules('nalog', 'ofdferma');
        $ofd_inn        = $this->diafan->configmodules('inn', 'ofdferma');
        $ofd_nds        = $this->diafan->configmodules('nds', 'ofdferma');

        return (
            $ofd_login &&
            $ofd_pass &&
            $ofd_nalog &&
            $ofd_inn &&
            $ofd_nds
        );
    }
    
    //Обновляем токен
    private function get_token(){
        
        if ( ! $this->check_settings()) {
            return $this->log_msg('Проверьте настройки, для корректной работы должны быть заполнены: Логин, Пароль, НДС по умолчанию, ИНН, Система налогообложения');
        };
        
        $ofd_login      = $this->diafan->configmodules('login', 'ofdferma');
        $ofd_pass       = $this->diafan->configmodules('password', 'ofdferma');
        $ofd_nalog      = $this->diafan->configmodules('nalog', 'ofdferma');
        $ofd_inn        = $this->diafan->configmodules('inn', 'ofdferma');
        $ofd_nds        = $this->diafan->configmodules('nds', 'ofdferma');
        
        $cache_meta = array(
            "login"     => $ofd_login,
            "pass"      => $ofd_pass,
            "nalog"     => $ofd_nalog,
            "inn"       => $ofd_inn,
            "nds"       => $ofd_nds
        );
        
        self::$ofdferma_token = $this->diafan->_cache->get(array_merge($cache_meta, array('name' => 'token')), self::MODULE_NAME);
        self::$ofdferma_token_exp_date = $this->diafan->_cache->get(array_merge($cache_meta, array('name' => 'token_exp_date')), self::MODULE_NAME);
        
        if (self::$ofdferma_token && (self::$ofdferma_token_exp_date > (time() - 10))) {
            return $this->log_msg('', true);
        }
        
        $ofd_auth_url  = self::AUTH_URL;

        $response = $this -> execute_request($ofd_auth_url, array(
            "Login"     => $ofd_login,
            "Password"  => $ofd_pass,
        ));
        
        if (isset($response['result'])){
            $result = $response['result'];
            
            //Обновляем токен
            self::$ofdferma_token = $result->Data->AuthToken;
            $this->diafan->_cache->save(self::$ofdferma_token, array_merge($cache_meta, array('name' => 'token')), self::MODULE_NAME);
                    
            self::$ofdferma_token_exp_date = $result->Data->ExpirationDateUtc;
            $this->diafan->_cache->save(self::$ofdferma_token_exp_date, array_merge($cache_meta, array('name' => 'token_exp_date')), self::MODULE_NAME);
            
            return $this->log_msg('', true);
        }
        
        return $response;
    }
    
    public function setDebug($val) {
        $this->_debug = $val;
    }
    
    public function getTextType($type) {
        $res = '';

        switch ($type) {
            case 'Income':
                $res = 'Приход';
                break;

            case 'IncomeReturn':
                $res = 'Возврат';
                break;
        }

        return $res;
    }
    
    private function getErrorByCode($code_id, $data) {
        $msg = '';

        SWITCH ($code_id) {
            case 1019:
                $msg = "Идентификатор счета '{$data['Request']['InvoiceId']}' уже существует в ОФД";
                break;
        }

        return $msg;
    }
    
    private function formatFloat( $int, $count = 2){
        return	round( (float)$int , $count);
    }
    
    private function customErrorHandler($errno, $errstr, $errfile, $errline, array $errcontext) {
        $this->log_msg($errstr);

        if (0 === error_reporting()) {
            return false;
        }
        throw new OfdfermaException($errstr, 0, $errno, $errfile, $errline);
    }
    
    public function log_msg($mess, $status = false, $params = array()) {
        
        if ( ! $status && $this->_debug && $mess) {
            file_put_contents(ABSOLUTE_PATH . 'modules/ofdferma/ofdferma.log', date('d.m.Y H:i:s') .':'. $mess . "\r\n", FILE_APPEND);
            
            $user_mail = $this->diafan->configmodules('email', 'ofdferma');
            if ($user_mail){  
                Custom::inc('includes/mail.php');

                send_mail($user_mail, 'OFD Ferma - Ошибка', $mess);
            }
        }
        
        return array_merge(array(
            'status'    => $status,
            'mess'      => $mess
        ), (array)$params);
    }
    
    public function getCheckLink($row){
        $check_link = self::CHECK_URL . "{$row['inn']}/{$row['RNM']}/{$row['FN']}/".ltrim($row['FDN'],0)."/".$row['FPD'];
        
        return $check_link;
    }
    
    /**
     * Функция для создания CURL-запроса
     *
     * @param $path
     * @param null|array $data
     * @return array
     */
    private function execute_request($url, $data = null) {  
        
        $options = array(
            "ssl" => array(
                "verify_peer" => false,
                "verify_peer_name" => false,
            ),
            'http' => array(
                'timeout' => 10,
                'ignore_errors' => true,
                'content' => json_encode($data),
                'header' => "Content-type: application/json\r\n" .
                "Accept: application/json" . "\r\n",
                "Content-Length: " . strlen(json_encode($data)) . "\r\n",
                'method' => 'POST',
            )
        );
        
        $context = stream_context_create($options);
        set_error_handler(array($this, 'customErrorHandler'));

        try {
            $result = file_get_contents($url, false, $context);
        } catch (Exception $e) {
            return $this->log_msg($e->getMessage());
        }
        
        restore_error_handler();
        $result = json_decode($result);
        
        if (isset($result->Status) && ($result->Status == 'Success')) {
            return $this->log_msg('', true, array('result' => $result));
        } else if (isset($result->Status) && ($result->Status == 'Failed')) {
            $msg = $result->Error->Message;
            if ($result->Error->Code) {
                if ($val = $this->getErrorByCode($result->Error->Code, $data)) {
                    $msg = $val;
                }
            }
            
            return $this->log_msg($msg, false, $result);
        } else {
            return $this->log_msg('some error', false, $result);
        }
    }
    
    //Обновляем статус чеков
    public function check_update($ids = array()){
        
        $res = $this->get_token();
        if ( ! $res['status']) {
            return $res;
        }
        
        $sWhere = '';
        if ($ids) {
            $aWhere = array();
            foreach ($ids as $id) {
                $aWhere[] = "'" . addslashes($id) . "'";
            }

            $sWhere = ' && `id` IN (' . implode(',', $aWhere) . ')';
        }

        $results = DB::query_fetch_all("SELECT * FROM {ofdferma_receipt} WHERE (`status` IS NULL OR `status` <> 'CONFIRMED' OR `status` <> 'FAILED')" . $sWhere);
        
        $this -> setDebug(1);
        
        foreach ($results as $result) {
            
            $data = array();
            $data['Request']['ReceiptId'] = $result['id'];
            
            //Отправляем запрос к ОФД
            $api_url = self::API_URL . "/status?AuthToken=" . self::$ofdferma_token;
            $response = $this -> execute_request($api_url, $data);

            if (isset($response['result'])){
                $this->update_receipt(array(
                    'status'            => $response['result']->Data->StatusName,
                    'status_message'    => $response['result']->Data->StatusMessage,
                    'FN'                => $response['result']->Data->Device->FN,
                    'RNM'               => $response['result']->Data->Device->RNM,
                    'FDN'               => $response['result']->Data->Device->FDN,
                    'FPD'               => $response['result']->Data->Device->FPD,
                ), $result['id']);
            }
        }
        
        $this -> setDebug(0);
    }
    
    /**
     * Обработка нового чека
     *
     * @param nc_payment_receipt $receipt
     */
    public function check($order_id, $type = 'Income') { 

        $temptype = $this -> getTextType($type);
        
        if ($this -> get_receipt($order_id, $type)){
            return $this->log_msg("Чек {$temptype} для заказа {$order_id} уже существует в модуле");
        }
        
        $res = $this->get_token();
        if ( ! $res['status']) {
            return $res;
        }
        
        $order = $this->getOrder($order_id);        
        
        //Формируем данные для запроса
        $prefix     = 'o';
        
        $ofd_nalog = $this->diafan->configmodules('nalog', 'ofdferma');
        $ofd_inn   = $this->diafan->configmodules('inn', 'ofdferma');
        $ofd_nds   = $this->diafan->configmodules('nds', 'ofdferma');
        
        //свертка
        $ofd_collapse       = $this->diafan->configmodules('collapse', 'ofdferma');
        $ofd_collapse_name  = $this->diafan->configmodules('collapse_name', 'ofdferma');
        
        
        $customer_phone = $this->getOrderField($order_id, 'phone');
        $customer_email = $this->getOrderField($order_id, 'email');

        
        $items = $this->getOrderPositions($order);
        if ( ! $items) {
            return $this->log_msg("В заказе #{$order_id} нет товаров");
        }

        //Формируем данные
        $data = array();
        $data['Request']['Inn'] = $ofd_inn;
        $data['Request']['Type'] = $type;
        $data['Request']['InvoiceId'] = $prefix . $order_id . '-' . $type;
        $data['Request']['LocalDate'] = date('Y-m-d\TH:i:s');
        $data['Request']['CustomerReceipt'] = array(
            'TaxationSystem'    => $ofd_nalog,
            'Email'             => $customer_email,
            'Phone'             => $customer_phone,
            'Items'             => array(),
        );

            
        // Get Items: Price / Sold / Email
        foreach ($items as $item) {
            $vat = $item['vat'];
            if ( ! $this -> checkNds($vat)){
                $vat = $ofd_nds;
            }
            
            array_push($data['Request']['CustomerReceipt']['Items'], array(
                    'Label'     => $item['name'],
                    'Price'     => $this->formatFloat( $item['price'] ),
                    'Quantity'  => $this->formatFloat( $item['count'], 3),
                    'Amount'    => $this->formatFloat( $item['total'] ),
                    'Vat'       => $vat,
                )
            );
        }
                
        //Если включена свертка
        if ($ofd_collapse) {
            $pos_name = $ofd_collapse_name ? $ofd_collapse_name : 'Undefined';

            $sum = 0;
            foreach ($data['Request']['CustomerReceipt']['Items'] as $item) {
                $sum += $item['Amount'];
            }
            $data['Request']['CustomerReceipt']['Items'] = array(array(
                    'Label'     => $pos_name,
                    'Price'     => $sum,
                    'Quantity'  => 1,
                    'Amount'    => $sum,
                    'Vat'       => $ofd_nds,
            ));
        }
        
        
        //Отправляем запрос к ОФД
        $api_url = self::API_URL . "/receipt?AuthToken=" . self::$ofdferma_token;
        $response = $this -> execute_request($api_url, $data);
        
        if (isset($response['result'])){
            $result = $response['result'];
            
            $this -> insert_receipt(array(
                'id'        => $result->Data->ReceiptId,
                'type'      => $data['Request']['Type'],
                'inn'       => $data['Request']['Inn'],
                'order_id'  => $order_id,
                'total'     => $order['summ']
            ));
                   
            return $this->log_msg("Чек {$temptype} для заказа {$order_id} сформирован: {$result->Data->ReceiptId}", true);
        }else{
            return $response;
        }
    }
}
